
data Coord = Position Double Double
           | Vector Double Double
           | Scalar Double
             deriving (Show, Eq)

add :: Coord -> Coord -> Coord
add (Position x y) (Vector dx dy) = Position (x + dx) (y + dy)
add (Vector dx dy) (Position x y) = Position (x + dx) (y + dy)
add (Vector dx1 dy1) (Vector dx2 dy2) = Vector (dx1 + dx2) (dy1 + dy2)
add (Scalar x) (Scalar y)         = Scalar (x + y)
add _ _ = error "undefined addition"

sub :: Coord -> Coord -> Coord
sub (Position x1 y1) (Position x2 y2) = Vector (x1 - x2) (y1 - y2)
sub (Position x y) (Vector dx dy)     = Position (x - dx) (y - dy)
sub (Scalar x) (Scalar y)         = Scalar (x - y)
sub _ _ = error "undefined subtraction"

mult :: Coord -> Coord -> Coord
mult (Position x y) (Position a b) = Scalar (x * a + y * b)
mult (Vector a b) (Vector x y) = Scalar (x * a + y * b)
mult (Vector dx dy) (Scalar c) = Vector (c * dx) (c * dy)
mult (Scalar c) (Vector dx dy) = Vector (c * dx) (c * dy)
mult (Scalar x) (Scalar y)     = Scalar (x * y)
mult _ _ = error "undefined multiplication"

neg :: Coord -> Coord
neg (Position x y) = Position (-x) (-y)
neg (Vector x y)   = Vector (-x) (-y)
neg (Scalar x)     = Scalar (-x)

pythagorean :: Double -> Double -> Double
pythagorean a b = sqrt (a * a + b * b)

abs' :: Coord -> Coord
abs' (Position x y) = Scalar $ pythagorean x y
abs' (Vector x y) = Scalar $ pythagorean x y
abs' (Scalar x)   = Scalar $ abs x

signum' :: Coord -> Coord
signum' (Scalar x) = Scalar $ signum x
signum' _ = error "undefined signum"

instance Num Coord where
    (+)    = add
    (-)    = sub
    (*)    = mult
    negate = neg
    abs    = abs'
    fromInteger = Scalar . fromInteger
    signum = signum'

